package com.tzld.piaoquan.ad.engine.service.score.strategy;

import com.ctrip.framework.apollo.spring.annotation.ApolloJsonValue;
import com.tzld.piaoquan.ad.engine.commons.dto.AdPlatformCreativeDTO;
import com.tzld.piaoquan.ad.engine.commons.param.RankRecommendRequestParam;
import com.tzld.piaoquan.ad.engine.commons.score.ScoreParam;
import com.tzld.piaoquan.ad.engine.commons.score.ScorerUtils;
import com.tzld.piaoquan.ad.engine.commons.thread.ThreadPoolFactory;
import com.tzld.piaoquan.ad.engine.commons.util.*;
import com.tzld.piaoquan.ad.engine.service.entity.FeatureContainer;
import com.tzld.piaoquan.ad.engine.service.entity.GuaranteeView;
import com.tzld.piaoquan.ad.engine.service.feature.Feature;
import com.tzld.piaoquan.recommend.feature.domain.ad.base.AdRankItem;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.collections4.MapUtils;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;
import org.xm.Similarity;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.stream.Collectors;

/**
 * 模型分校准实验
 */
@Slf4j
@Service
public class RankStrategyBy679 extends RankStrategyBasic {


    private Map<String, double[]> bucketsMap = new HashMap<>();

    private Map<String, Double> bucketsLen = new HashMap<>();

    @Value("${word2vec.exp:694}")
    private String word2vecExp;

    @ApolloJsonValue("${rank.score.weight.679:{}}")
    private Map<String, Double> weightMap;

    @Override
    public List<AdRankItem> adItemRank(RankRecommendRequestParam request, ScoreParam scoreParam) {

        Map<String, Double> weightParam = ObjUtil.nullOrDefault(weightMap, new HashMap<>());

        Set<String> noApiAdVerIds = getNoApiAdVerIds();

        Map<Long, Double> creativeScoreCoefficient = getCreativeScoreCoefficient();

        long ts = System.currentTimeMillis() / 1000;

        String brand = scoreParam.getRequestContext().getMachineinfoBrand();
        if (StringUtils.isNotEmpty(brand)) {
            scoreParam.getRequestContext().setMachineinfoBrand(brand + "-n");
        }

        long start = System.currentTimeMillis();
        // 特征处理
        FeatureContainer featureContainer = this.extractAndProcessFeatures(scoreParam, request, ts);

        Map<String, String> userFeatureMap = new HashMap<>();
        List<TupleMapEntry<Tuple5>> midActionList = this.handleC1Feature(featureContainer.getC1Feature(), userFeatureMap);
        Map<String, Double> midTimeDiffMap = this.parseC1FeatureListToTimeDiffMap(midActionList, ts);
        Map<String, Double> actionStaticMap = this.parseC1FeatureListToActionStaticMap(midActionList);
        Map<String, Map<String, Double>> vidRankMaps = this.parseD2FeatureMap(featureContainer.getD2Feature());
        long time1 = System.currentTimeMillis();
        Map<String, GuaranteeView> map = getGuaranteeViewMap(request, scoreParam);
        List<AdRankItem> adRankItems = new ArrayList<>();
        Random random = new Random();
        List<Future<AdRankItem>> futures = new ArrayList<>();
        CountDownLatch cdl1 = new CountDownLatch(request.getAdIdList().size());
        for (AdPlatformCreativeDTO dto : request.getAdIdList()) {
            Future<AdRankItem> future = ThreadPoolFactory.feature().submit(() -> {
                AdRankItem adRankItem = new AdRankItem();
                try {
                    adRankItem.setAdId(dto.getCreativeId());
                    adRankItem.setCreativeCode(dto.getCreativeCode());
                    adRankItem.setAdVerId(dto.getAdVerId());
                    adRankItem.setVideoId(request.getVideoId());
                    adRankItem.setCpa(dto.getCpa());
                    adRankItem.setId(dto.getAdId());
                    adRankItem.setCampaignId(dto.getCampaignId());
                    adRankItem.setCpm(ObjUtil.nullOrDefault(dto.getCpm(), 90).doubleValue());
                    adRankItem.setSkuId(dto.getSkuId());
                    adRankItem.setRandom(random.nextInt(1000));
                    if (noApiAdVerIds.contains(dto.getAdVerId())) {
                        adRankItem.getExt().put("isApi", "0");
                    } else {
                        adRankItem.getExt().put("isApi", "1");
                    }
                    adRankItem.getExt().put("recallsources", dto.getRecallSources());
                    setGuaranteeWeight(map, dto.getAdVerId(), adRankItem.getExt());
                    String cidStr = dto.getCreativeId().toString();
                    Map<String, String> cidFeatureMap = adRankItem.getFeatureMap();
                    Map<String, Map<String, String>> cidFeature = featureContainer.getCidFeature().getOrDefault(cidStr, new HashMap<>());
                    Map<String, String> b1Feature = cidFeature.getOrDefault("alg_cid_feature_basic_info", new HashMap<>());

                    Map<String, Map<String, String>> adVerFeature = featureContainer.getAdVerFeature().getOrDefault(dto.getAdVerId(), new HashMap<>());

                    Map<String, String> d1Feature = cidFeature.getOrDefault("alg_cid_feature_vid_cf", new HashMap<>());

                    this.handleB1Feature(b1Feature, cidFeatureMap, cidStr);
                    this.handleB2ToB5AndB8ToB9Feature(cidFeature, adVerFeature, cidFeatureMap);
                    this.handleB6ToB7Feature(cidFeature, cidFeatureMap);
                    this.handleC1UIFeature(midTimeDiffMap, actionStaticMap, cidFeatureMap, cidStr);
                    this.handleD1Feature(d1Feature, cidFeatureMap);
                    this.handleD2Feature(vidRankMaps, cidFeatureMap, cidStr);
                    return adRankItem;
                } finally {
                    cdl1.countDown();
                }
            });
            futures.add(future);
        }
        try {
            cdl1.await(300, TimeUnit.MILLISECONDS);
        } catch (Exception e) {
            log.error("handleE1AndE2Feature and handleD3AndB1Feature wait timeout", e);
        }
        for (Future<AdRankItem> future : futures) {
            try {
                if (future.isDone()) {
                    adRankItems.add(future.get());
                }
            } catch (Exception e) {
                log.error("Feature handle error", e);
            }
        }

        long time2 = System.currentTimeMillis();
        // feature3
        CountDownLatch cdl2 = new CountDownLatch(adRankItems.size() * 2);
        for (AdRankItem item : adRankItems) {
            String cidStr = String.valueOf(item.getAdId());
            Map<String, Map<String, String>> cidFeature = featureContainer.getCidFeature().getOrDefault(cidStr, new HashMap<>());
            Map<String, String> b1Feature = cidFeature.getOrDefault("alg_cid_feature_basic_info", new HashMap<>());
            String title = b1Feature.getOrDefault("cidtitle", "");
            ThreadPoolFactory.defaultPool().submit(() -> {
                try {
                    this.handleE1AndE2Feature(featureContainer.getE1Feature(), featureContainer.getE2Feature(), title, item.getFeatureMap(), scoreParam);
                } finally {
                    cdl2.countDown();
                }
            });
            ThreadPoolFactory.defaultPool().submit(() -> {
                try {
                    this.handleD3AndB1Feature(featureContainer.getD3Feature(), title, item.getFeatureMap(), scoreParam);
                } finally {
                    cdl2.countDown();
                }
            });
        }
        long time31 = System.currentTimeMillis();
        try {
            cdl2.await(150, TimeUnit.MILLISECONDS);
        } catch (Exception e) {
            log.error("handleE1AndE2Feature and handleD3AndB1Feature wait timeout", e);
        }

        // feature4
        long time3 = System.currentTimeMillis();
        // 分桶
        this.readBucketFile();
        userFeatureMap = this.featureBucket(userFeatureMap);
        CountDownLatch cdl4 = new CountDownLatch(adRankItems.size());
        for (AdRankItem adRankItem : adRankItems) {
            ThreadPoolFactory.feature().submit(() -> {
                try {
                    Map<String, String> featureMap = adRankItem.getFeatureMap();
                    adRankItem.setFeatureMap(this.featureBucket(featureMap));
                } finally {
                    cdl4.countDown();
                }
            });
        }
        try {
            cdl4.await(100, TimeUnit.MILLISECONDS);
        } catch (Exception e) {
            log.error("handleE1AndE2Feature and handleD3AndB1Feature wait timeout", e);
        }
        long time4 = System.currentTimeMillis();
        // 打分排序
        // getScorerPipeline
        List<AdRankItem> result = ScorerUtils.getScorerPipeline(ScorerUtils.XGBOOST_SCORE_CONF_20241105).scoring(featureContainer.getSceneFeatureMap(), userFeatureMap, adRankItems);
        long time5 = System.currentTimeMillis();
        // loop
        double cpmCoefficient = weightParam.getOrDefault("cpmCoefficient", 0.9);
        for (AdRankItem item : result) {

            double scoreCoefficient = creativeScoreCoefficient.getOrDefault(item.getAdId(), 1d);
            double guaranteeScoreCoefficient = getGuaranteeScoreCoefficient(scoreParam, item.getExt());
            item.setScore(item.getLrScore() * scoreCoefficient * item.getCpa() * guaranteeScoreCoefficient);
            item.getScoreMap().put("guaranteeScoreCoefficient", guaranteeScoreCoefficient);
            item.getScoreMap().put("cpa", item.getCpa());
            item.getScoreMap().put("cpm", item.getCpm());
            item.getScoreMap().put("cpmCoefficient", cpmCoefficient);
            item.getScoreMap().put("scoreCoefficient", scoreCoefficient);
            item.getFeatureMap().putAll(userFeatureMap);
            item.getFeatureMap().putAll(featureContainer.getSceneFeatureMap());

            // 没有转化回传的广告主，使用后台配置的CPM
            if (noApiAdVerIds.contains(item.getAdVerId())) {
                item.setScore(item.getCpm() * cpmCoefficient / 1000);
            }

            setMetaFeatureMap(featureContainer, item);
        }

        log.info("cost={}, feature1={}, feature2={}, feature31={}, feature32={}, feature4={}, getScorerPipeline={}, " +
                        "adIdSize={}, adRankItemsSize={}",
                time5 - start, time1 - start, time2 - time1, time31 - time2, time3 - time31, time4 - time3,
                time5 - time4, request.getAdIdList().size(), adRankItems.size());

        result.sort(ComparatorUtil.equalsRandomComparator());

        return result;
    }


    protected void handleB2ToB5AndB8ToB9Feature(Map<String, Map<String, String>> c1Feature, Map<String, Map<String, String>> adVerFeature, Map<String, String> cidFeatureMap) {
        Map<String, String> b2Feature = adVerFeature.getOrDefault("alg_cid_feature_adver_action", new HashMap<>());
        Map<String, String> b3Feature = c1Feature.getOrDefault("alg_cid_feature_cid_action", new HashMap<>());
        Map<String, String> b4Feature = c1Feature.getOrDefault("alg_cid_feature_region_action", new HashMap<>());
        Map<String, String> b5Feature = c1Feature.getOrDefault("alg_cid_feature_app_action", new HashMap<>());
        Map<String, String> b8Feature = c1Feature.getOrDefault("alg_cid_feature_brand_action", new HashMap<>());
        Map<String, String> b9Feature = c1Feature.getOrDefault("alg_cid_feature_weChatVersion_action", new HashMap<>());

        List<String> timeList = Arrays.asList("1h", "2h", "3h", "6h", "12h", "1d", "3d", "7d", "yesterday", "today");
        List<Tuple2<Map<String, String>, String>> featureList = Arrays.asList(
                new Tuple2<>(b2Feature, "b2"),
                new Tuple2<>(b3Feature, "b3"),
                new Tuple2<>(b4Feature, "b4"),
                new Tuple2<>(b5Feature, "b5"),
                new Tuple2<>(b8Feature, "b8"),
                new Tuple2<>(b9Feature, "b9")
        );
        for (Tuple2<Map<String, String>, String> tuple2 : featureList) {
            Map<String, String> feature = tuple2.f1;
            String prefix = tuple2.f2;
            for (String time : timeList) {
                double view = Double.parseDouble(feature.getOrDefault("ad_view_" + time, "0"));
                double click = Double.parseDouble(feature.getOrDefault("ad_click_" + time, "0"));
                double conver = Double.parseDouble(feature.getOrDefault("ad_conversion_" + time, "0"));
                double income = Double.parseDouble(feature.getOrDefault("ad_income_" + time, "0"));
                double f2 = NumUtil.div(conver, view);
                double ecpm = NumUtil.div(income * 1000, view);
                cidFeatureMap.put(prefix + "_" + time + "_ctr", String.valueOf(NumUtil.div(click, view)));
                cidFeatureMap.put(prefix + "_" + time + "_ctcvr", String.valueOf(f2));
                cidFeatureMap.put(prefix + "_" + time + "_cvr", String.valueOf(NumUtil.div(conver, click)));
                cidFeatureMap.put(prefix + "_" + time + "_conver", String.valueOf(conver));
                cidFeatureMap.put(prefix + "_" + time + "_ecpm", String.valueOf(ecpm));

                cidFeatureMap.put(prefix + "_" + time + "_click", String.valueOf(click));
                cidFeatureMap.put(prefix + "_" + time + "_conver*log(view)", String.valueOf(conver * NumUtil.log(view)));
                cidFeatureMap.put(prefix + "_" + time + "_conver*ctcvr", String.valueOf(conver * f2));
            }
        }

    }

    protected void handleB6ToB7Feature(Map<String, Map<String, String>> c1Feature, Map<String, String> cidFeatureMap) {
        Map<String, String> b6Feature = c1Feature.getOrDefault("alg_cid_feature_week_action", new HashMap<>());
        Map<String, String> b7Feature = c1Feature.getOrDefault("alg_cid_feature_hour_action", new HashMap<>());

        List<String> timeList = Arrays.asList("7d", "14d");
        List<Tuple2<Map<String, String>, String>> featureList = Arrays.asList(
                new Tuple2<>(b6Feature, "b6"),
                new Tuple2<>(b7Feature, "b7")
        );
        for (Tuple2<Map<String, String>, String> tuple2 : featureList) {
            Map<String, String> feature = tuple2.f1;
            String prefix = tuple2.f2;
            for (String time : timeList) {
                double view = Double.parseDouble(feature.getOrDefault("ad_view_" + time, "0"));
                double click = Double.parseDouble(feature.getOrDefault("ad_click_" + time, "0"));
                double conver = Double.parseDouble(feature.getOrDefault("ad_conversion_" + time, "0"));
                double income = Double.parseDouble(feature.getOrDefault("ad_income_" + time, "0"));
                double f2 = NumUtil.div(conver, view);
                double ecpm = NumUtil.div(income * 1000, view);
                cidFeatureMap.put(prefix + "_" + time + "_ctr", String.valueOf(NumUtil.div(click, view)));
                cidFeatureMap.put(prefix + "_" + time + "_ctcvr", String.valueOf(f2));
                cidFeatureMap.put(prefix + "_" + time + "_cvr", String.valueOf(NumUtil.div(conver, click)));
                cidFeatureMap.put(prefix + "_" + time + "_conver", String.valueOf(conver));
                cidFeatureMap.put(prefix + "_" + time + "_ecpm", String.valueOf(ecpm));

                cidFeatureMap.put(prefix + "_" + time + "_click", String.valueOf(click));
                cidFeatureMap.put(prefix + "_" + time + "_conver*log(view)", String.valueOf(conver * NumUtil.log(view)));
                cidFeatureMap.put(prefix + "_" + time + "_conver*ctcvr", String.valueOf(conver * f2));
            }
        }

    }


    protected void handleD1Feature(Map<String, String> d1Feature, Map<String, String> featureMap) {
        for (String prefix : Arrays.asList("3h", "6h", "12h", "1d", "3d", "7d")) {
            double view = Double.parseDouble(d1Feature.getOrDefault("ad_view_" + prefix, "0"));
            double click = Double.parseDouble(d1Feature.getOrDefault("ad_click_" + prefix, "0"));
            double conver = Double.parseDouble(d1Feature.getOrDefault("ad_conversion_" + prefix, "0"));
            double income = Double.parseDouble(d1Feature.getOrDefault("ad_income_" + prefix, "0"));
            featureMap.put("d1_feature_" + prefix + "_ctr", String.valueOf(NumUtil.div(click, view)));
            featureMap.put("d1_feature_" + prefix + "_ctcvr", String.valueOf(NumUtil.div(conver, view)));
            featureMap.put("d1_feature_" + prefix + "_cvr", String.valueOf(NumUtil.div(conver, click)));
            featureMap.put("d1_feature_" + prefix + "_conver", String.valueOf(conver));
            featureMap.put("d1_feature_" + prefix + "_ecpm", String.valueOf(NumUtil.div(income * 1000, view)));
        }
    }

    protected void handleE1AndE2Feature(Map<String, String> e1Feature, Map<String, String> e2Feature, String title,
                                        Map<String, String> featureMap, ScoreParam scoreParam) {
        if (StringUtils.isEmpty(title)) {
            return;
        }

        List<Tuple2<Map<String, String>, String>> tuple2List = Arrays.asList(new Tuple2<>(e1Feature, "e1"), new Tuple2<>(e2Feature, "e2"));

        List<String> tagsFieldList = Arrays.asList("tags_3d", "tags_7d", "tags_14d");
        for (Tuple2<Map<String, String>, String> tuple2 : tuple2List) {
            Map<String, String> feature = tuple2.f1;
            String prefix = tuple2.f2;
            if (MapUtils.isEmpty(feature)) {
                continue;
            }

            for (String tagsField : tagsFieldList) {
                if (StringUtils.isNotEmpty(feature.get(tagsField))) {
                    String tags = feature.get(tagsField);
                    Double[] doubles;
                    if (scoreParam.getExpCodeSet().contains(word2vecExp)) {
                        doubles = ExtractorUtils.funcC34567ForTagsNew(tags, title);
                    } else {
                        doubles = ExtractorUtils.funcC34567ForTags(tags, title);
                    }
                    featureMap.put(prefix + "_" + tagsField + "_matchnum", String.valueOf(doubles[0]));
                    featureMap.put(prefix + "_" + tagsField + "_maxscore", String.valueOf(doubles[1]));
                    featureMap.put(prefix + "_" + tagsField + "_avgscore", String.valueOf(doubles[2]));
                }
            }
        }
    }

    private void readBucketFile() {
        if (MapUtils.isNotEmpty(bucketsMap)) {
            return;
        }
        synchronized (this) {
            InputStream resourceStream = RankStrategyBy679.class.getClassLoader().getResourceAsStream("20240718_ad_bucket_688.txt");
            if (resourceStream != null) {
                try (BufferedReader reader = new BufferedReader(new InputStreamReader(resourceStream))) {
                    Map<String, double[]> bucketsMap = new HashMap<>();
                    Map<String, Double> bucketsLen = new HashMap<>();
                    String line;
                    while ((line = reader.readLine()) != null) {
                        // 替换空格和换行符，过滤空行
                        line = line.replace(" ", "").replaceAll("\n", "");
                        if (!line.isEmpty()) {
                            String[] rList = line.split("\t");
                            if (rList.length == 3) {
                                String key = rList[0];
                                double value1 = Double.parseDouble(rList[1]);
                                bucketsLen.put(key, value1);
                                double[] value2 = Arrays.stream(rList[2].split(",")).mapToDouble(Double::valueOf).toArray();
                                bucketsMap.put(key, value2);
                            }
                        }
                    }
                    this.bucketsMap = bucketsMap;
                    this.bucketsLen = bucketsLen;
                } catch (IOException e) {
                    log.error("something is wrong in parse bucket file:", e);
                }
            } else {
                log.error("no bucket file");
            }
        }
    }

    private Map<String, String> featureBucket(Map<String, String> featureMap) {
        Map<String, String> newFeatureMap = new ConcurrentHashMap<>(featureMap.size());
        for (Map.Entry<String, String> entry : featureMap.entrySet()) {
            String name = entry.getKey();
            double score = Double.parseDouble(entry.getValue());
            // 注意：0值、不在分桶文件中的特征，会被过滤掉。
            if (score > 1E-8) {
                if (this.bucketsMap.containsKey(name) && this.bucketsLen.containsKey(name)) {
                    double[] buckets = this.bucketsMap.get(name);
                    double bucketNum = this.bucketsLen.get(name);
                    Double scoreNew = 1.0 / bucketNum * (ExtractorUtils.findInsertPosition(buckets, score) + 1.0);
                    newFeatureMap.put(name, String.valueOf(scoreNew));
                } else {
                    newFeatureMap.put(name, String.valueOf(score));
                }
            }
        }

        return newFeatureMap;
    }

}
