package com.tzld.piaoquan.ad.engine.service.score.strategy;

import com.alibaba.fastjson.JSONObject;
import com.ctrip.framework.apollo.spring.annotation.ApolloJsonValue;
import com.tzld.piaoquan.ad.engine.commons.redis.AdRedisHelper;
import com.tzld.piaoquan.ad.engine.commons.score.ScoreParam;
import com.tzld.piaoquan.ad.engine.commons.score.ScorerUtils;
import com.tzld.piaoquan.ad.engine.commons.thread.ThreadPoolFactory;
import com.tzld.piaoquan.ad.engine.commons.util.*;
import com.tzld.piaoquan.ad.engine.service.entity.FeatureContainer;
import com.tzld.piaoquan.ad.engine.service.entity.GuaranteeView;
import com.tzld.piaoquan.ad.engine.service.feature.Feature;
import com.tzld.piaoquan.ad.engine.commons.dto.AdPlatformCreativeDTO;
import com.tzld.piaoquan.ad.engine.commons.param.RankRecommendRequestParam;
import com.tzld.piaoquan.recommend.feature.domain.ad.base.AdRankItem;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.collections4.MapUtils;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Component;
import org.xm.Similarity;

import javax.annotation.PostConstruct;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.stream.Collectors;

import static com.tzld.piaoquan.ad.engine.commons.math.Const.*;

@Slf4j
@Component
public class RankStrategyBy688 extends RankStrategyBasic {

    private Map<String, double[]> bucketsMap = new HashMap<>();

    private Map<String, Double> bucketsLen = new HashMap<>();

    @Value("${word2vec.exp:694}")
    private String word2vecExp;


    // FIXME(zhoutian): 可能需要独立配置
    @ApolloJsonValue("${rank.score.weight.680:{}}")
    private Map<String, Double> weightMap;

    @ApolloJsonValue("${rank.score.neg_sample_rate:0.01}")
    Double negSampleRate;

    Set<String> sparseFeatureSet;


    @PostConstruct
    public void afterInit() {
        this.readBucketFile();
        this.initSparseFeatureNames();
    }


    @Override
    public List<AdRankItem> adItemRank(RankRecommendRequestParam request, ScoreParam scoreParam) {
        Map<String, Double> weightParam = ObjUtil.nullOrDefault(weightMap, new HashMap<>());


        Map<Long, Double> creativeScoreCoefficient = getCreativeScoreCoefficient();
        Set<String> noApiAdVerIds = getNoApiAdVerIds();

        long ts = System.currentTimeMillis() / 1000;

        String brand = scoreParam.getRequestContext().getMachineinfoBrand();
        if (StringUtils.isNotEmpty(brand)) {
            scoreParam.getRequestContext().setMachineinfoBrand(brand + "-n");
        }

        long start = System.currentTimeMillis();
        // 特征处理
        FeatureContainer featureContainer = this.extractAndProcessFeatures(scoreParam, request, ts);

        Map<String, String> userFeatureMap = new HashMap<>();
        List<TupleMapEntry<Tuple5>> midActionList = this.handleC1Feature(featureContainer.getC1Feature(), userFeatureMap);
        Map<String, Double> midTimeDiffMap = this.parseC1FeatureListToTimeDiffMap(midActionList, ts);
        Map<String, Double> actionStaticMap = this.parseC1FeatureListToActionStaticMap(midActionList);
        Map<String, Map<String, Double>> vidRankMaps = this.parseD2FeatureMap(featureContainer.getD2Feature());

        Map<String, String> reqFeature = featureContainer.getReqFeature();
        userFeatureMap.put("brand", reqFeature.getOrDefault("brand", ""));
        userFeatureMap.put("region", reqFeature.getOrDefault("region", ""));
        userFeatureMap.put("city", reqFeature.getOrDefault("city", ""));
        userFeatureMap.put("vid", reqFeature.getOrDefault("vid", ""));
        userFeatureMap.put("apptype", reqFeature.getOrDefault("apptype", ""));
        userFeatureMap.put("is_first_layer", reqFeature.getOrDefault("is_first_layer", ""));
        userFeatureMap.put("root_source_scene", reqFeature.getOrDefault("root_source_scene", ""));
        userFeatureMap.put("root_source_channel", reqFeature.getOrDefault("root_source_channel", ""));

        Map<String, String> d3Feature = featureContainer.getD3Feature();
        Map<String, String> e1Feature = featureContainer.getE1Feature();
        Map<String, String> e2Feature = featureContainer.getE2Feature();
        Map<String, String> g1Feature = featureContainer.getG1Feature();
        Map<String, String> g2Feature = featureContainer.getG2Feature();
        userFeatureMap.put("cate1", d3Feature.get("merge_first_level_cate"));
        userFeatureMap.put("cate2", d3Feature.get("merge_second_level_cate"));
        userFeatureMap.put("user_vid_return_tags_2h", e1Feature.getOrDefault("tags_2h", null));
        userFeatureMap.put("user_vid_return_tags_1d", e1Feature.getOrDefault("tags_1d", null));
        userFeatureMap.put("user_vid_return_tags_3d", e1Feature.getOrDefault("tags_3d", null));
        userFeatureMap.put("user_vid_return_tags_7d", e1Feature.getOrDefault("tags_7d", null));
        userFeatureMap.put("user_vid_return_tags_14d", e1Feature.getOrDefault("tags_14d", null));
        userFeatureMap.put("title_split", d3Feature.getOrDefault("title_split", null));
        userFeatureMap.put("user_vid_share_tags_1d", e2Feature.getOrDefault("tags_1d", null));
        userFeatureMap.put("user_vid_share_tags_14d", e2Feature.getOrDefault("tags_14d", null));
        userFeatureMap.put("user_vid_return_cate1_14d", g1Feature.getOrDefault("cate1_14d", null));
        userFeatureMap.put("user_vid_return_cate2_14d", g1Feature.getOrDefault("cate2_14d", null));
        userFeatureMap.put("user_vid_share_cate1_14d", g2Feature.getOrDefault("cate1_14d", null));
        userFeatureMap.put("user_vid_share_cate2_14d", g2Feature.getOrDefault("cate2_14d", null));
        long time1 = System.currentTimeMillis();

        Map<String, GuaranteeView> map = getGuaranteeViewMap(request, scoreParam);
        List<AdRankItem> adRankItems = new ArrayList<>();
        Random random = new Random();
        List<Future<AdRankItem>> futures = new ArrayList<>();
        CountDownLatch cdl1 = new CountDownLatch(request.getAdIdList().size());
        for (AdPlatformCreativeDTO dto : request.getAdIdList()) {
            Future<AdRankItem> future = ThreadPoolFactory.feature().submit(() -> {
                AdRankItem adRankItem = new AdRankItem();
                try {
                    adRankItem.setAdId(dto.getCreativeId());
                    adRankItem.setCreativeCode(dto.getCreativeCode());
                    adRankItem.setAdVerId(dto.getAdVerId());
                    adRankItem.setVideoId(request.getVideoId());
                    adRankItem.setCpa(dto.getCpa());
                    adRankItem.setId(dto.getAdId());
                    adRankItem.setCampaignId(dto.getCampaignId());
                    adRankItem.setCpm(ObjUtil.nullOrDefault(dto.getCpm(), 90).doubleValue());
                    adRankItem.setSkuId(dto.getSkuId());
                    adRankItem.setRandom(random.nextInt(1000));
                    if (noApiAdVerIds.contains(dto.getAdVerId())) {
                        adRankItem.getExt().put("isApi", "0");
                    } else {
                        adRankItem.getExt().put("isApi", "1");
                    }
                    adRankItem.getExt().put("recallsources", dto.getRecallSources());
                    setGuaranteeWeight(map, dto.getAdVerId(), adRankItem.getExt());
                    String cidStr = dto.getCreativeId().toString();
                    Map<String, String> cidFeatureMap = adRankItem.getFeatureMap();
                    Map<String, Map<String, String>> cidFeature = featureContainer.getCidFeature().getOrDefault(cidStr, new HashMap<>());
                    Map<String, String> b1Feature = cidFeature.getOrDefault("alg_cid_feature_basic_info", new HashMap<>());

                    Map<String, Map<String, String>> adVerFeature = featureContainer.getAdVerFeature().getOrDefault(dto.getAdVerId(), new HashMap<>());
                    Map<String, Map<String, String>> skuFeature = featureContainer.getSkuFeature().getOrDefault(String.valueOf(dto.getSkuId()), new HashMap<>());
                    Map<String, String> d1Feature = cidFeature.getOrDefault("alg_cid_feature_vid_cf", new HashMap<>());

                    this.handleB1Feature(b1Feature, cidFeatureMap, cidStr);
                    this.handleB2ToB5AndB8ToB9Feature(cidFeature, adVerFeature, cidFeatureMap);
                    this.handleB6ToB7Feature(cidFeature, cidFeatureMap);
                    this.handleC1UIFeature(midTimeDiffMap, actionStaticMap, cidFeatureMap, cidStr);
                    this.handleD1Feature(d1Feature, cidFeatureMap);
                    this.handleD2Feature(vidRankMaps, cidFeatureMap, cidStr);
                    this.handleH1AndH2Feature(skuFeature, adVerFeature, cidFeatureMap);

                    cidFeatureMap.put("cid", dto.getCreativeId() != null ? String.valueOf(dto.getCreativeId()) : "");
                    cidFeatureMap.put("adid", dto.getAdId() != null ? String.valueOf(dto.getAdId()) : "");
                    cidFeatureMap.put("adverid", dto.getAdVerId() != null ? dto.getAdVerId() : "");
                    cidFeatureMap.put("profession", dto.getProfession() != null ? dto.getProfession() : "");
                    return adRankItem;
                } finally {
                    cdl1.countDown();
                }
            });
            futures.add(future);
        }
        try {
            cdl1.await(300, TimeUnit.MILLISECONDS);
        } catch (Exception e) {
            log.error("handleE1AndE2Feature and handleD3AndB1Feature wait timeout", e);
        }
        for (Future<AdRankItem> future : futures) {
            try {
                if (future.isDone()) {
                    adRankItems.add(future.get());
                }
            } catch (Exception e) {
                log.error("Feature handle error", e);
            }
        }

        long time2 = System.currentTimeMillis();
        // feature3
        CountDownLatch cdl2 = new CountDownLatch(adRankItems.size() * 2);
        for (AdRankItem item : adRankItems) {
            String cidStr = String.valueOf(item.getAdId());
            Map<String, Map<String, String>> cidFeature = featureContainer.getCidFeature().getOrDefault(cidStr, new HashMap<>());
            Map<String, String> b1Feature = cidFeature.getOrDefault("alg_cid_feature_basic_info", new HashMap<>());
            String title = b1Feature.getOrDefault("cidtitle", "");
            ThreadPoolFactory.defaultPool().submit(() -> {
                try {
                    this.handleE1AndE2Feature(e1Feature, e2Feature, title, item.getFeatureMap(), scoreParam);
                } finally {
                    cdl2.countDown();
                }
            });
            ThreadPoolFactory.defaultPool().submit(() -> {
                try {
                    this.handleD3AndB1Feature(d3Feature, title, item.getFeatureMap(), scoreParam);
                } finally {
                    cdl2.countDown();
                }
            });
        }
        try {
            cdl2.await(150, TimeUnit.MILLISECONDS);
        } catch (Exception e) {
            log.error("handleE1AndE2Feature and handleD3AndB1Feature wait timeout", e);
        }

        long time3 = System.currentTimeMillis();
        // 分桶
        userFeatureMap = this.featureBucket(userFeatureMap);
        CountDownLatch cdl4 = new CountDownLatch(adRankItems.size());
        for (AdRankItem adRankItem : adRankItems) {
            ThreadPoolFactory.feature().submit(() -> {
                try {
                    Map<String, String> featureMap = adRankItem.getFeatureMap();
                    adRankItem.setFeatureMap(this.featureBucket(featureMap));
                } finally {
                    cdl4.countDown();
                }
            });
        }
        try {
            cdl4.await(100, TimeUnit.MILLISECONDS);
        } catch (Exception e) {
            log.error("handleE1AndE2Feature and handleD3AndB1Feature wait timeout", e);
        }
        long time4 = System.currentTimeMillis();
        // 打分排序
        // getScorerPipeline
        List<AdRankItem> result = ScorerUtils.getScorerPipeline(ScorerUtils.PAI_SCORE_CONF_20250214)
                .scoring(featureContainer.getSceneFeatureMap(), userFeatureMap, adRankItems);
        long time5 = System.currentTimeMillis();

        // calibrate score for negative sampling
        for (AdRankItem item : result) {
            double originalScore = item.getLrScore();
            double calibratedScore = originalScore / (originalScore + (1 - originalScore) / negSampleRate);
            item.setLrScore(calibratedScore);
            item.getScoreMap().put("originCtcvrScore", originalScore);
            item.getScoreMap().put("ctcvrScore", calibratedScore);
        }

        // loop
        double cpmCoefficient = weightParam.getOrDefault("cpmCoefficient", 0.9);

        for (AdRankItem item : result) {

            double scoreCoefficient = creativeScoreCoefficient.getOrDefault(item.getAdId(), 1d);
            double guaranteeScoreCoefficient = getGuaranteeScoreCoefficient(scoreParam, item.getExt());
            item.setScore(item.getLrScore() * scoreCoefficient * item.getCpa() * guaranteeScoreCoefficient);
            item.getScoreMap().put("guaranteeScoreCoefficient", guaranteeScoreCoefficient);
            item.getScoreMap().put("cpa", item.getCpa());
            item.getScoreMap().put("cpm", item.getCpm());
            item.getScoreMap().put("cpmCoefficient", cpmCoefficient);
            item.getScoreMap().put("scoreCoefficient", scoreCoefficient);
            item.getFeatureMap().putAll(userFeatureMap);
            item.getFeatureMap().putAll(featureContainer.getSceneFeatureMap());

            // 没有转化回传的广告主，使用后台配置的CPM
            if (noApiAdVerIds.contains(item.getAdVerId())) {
                item.setScore(item.getCpm() * cpmCoefficient / 1000);
            }
        }

        result.sort(ComparatorUtil.equalsRandomComparator());

        if (CollectionUtils.isNotEmpty(result)) {
            AdRankItem top1Item = result.get(0);
            setMetaFeatureMap(featureContainer, top1Item);
        }
        long time6 = System.currentTimeMillis();
        log.info("cost={}, getFeature={}, handleFeature={},  similar={}, bucketFeature={}, getScorerPipeline={}, " +
                        "other={}, adIdSize={}, adRankItemsSize={}",
                time6 - start, time1 - start, time2 - time1, time3 - time2, time4 - time3,
                time5 - time4, time6 - time5, request.getAdIdList().size(), adRankItems.size());

        return result;
    }


    protected List<TupleMapEntry<Tuple5>> handleC1Feature(Map<String, String> c1Feature, Map<String, String> featureMap) {

        //用户近1年内是否有转化
        if (c1Feature.containsKey("user_has_conver_1y") && c1Feature.get("user_has_conver_1y") != null) {
            featureMap.put("user_has_conver_1y", c1Feature.get("user_has_conver_1y"));
        }

        // 用户特征
        List<TupleMapEntry<Tuple5>> midActionList = new ArrayList<>();
        if (c1Feature.containsKey("action")) {
            String action = c1Feature.get("action");
            midActionList = Arrays.stream(action.split(","))
                    .map(r -> {
                        String[] rList = r.split(":");
                        Tuple5 tuple5 = new Tuple5(rList[1], rList[2], rList[3], rList[4], rList[5]);
                        return new TupleMapEntry<>(rList[0], tuple5);
                    })
                    // TODO 倒排
                    .sorted((a, b) -> Integer.compare(Integer.parseInt(b.value.f1), Integer.parseInt(a.value.f1)))
                    .collect(Collectors.toList());
        }

        double viewAll = midActionList.size();
        double clickAll = midActionList.stream().mapToInt(e -> Integer.parseInt(e.value.f2)).sum();
        double converAll = midActionList.stream().mapToInt(e -> Integer.parseInt(e.value.f3)).sum();
        double incomeAll = midActionList.stream().mapToInt(e -> Integer.parseInt(e.value.f4)).sum();
        featureMap.put("viewAll", String.valueOf(viewAll));
        featureMap.put("clickAll", String.valueOf(clickAll));
        featureMap.put("converAll", String.valueOf(converAll));
        featureMap.put("incomeAll", String.valueOf(incomeAll));
        featureMap.put("ctr_all", String.valueOf(NumUtil.div(clickAll, viewAll)));
        featureMap.put("ctcvr_all", String.valueOf(NumUtil.div(converAll, viewAll)));
        featureMap.put("cvr_all", String.valueOf(NumUtil.div(clickAll, converAll)));
        featureMap.put("ecpm_all", String.valueOf(NumUtil.div(incomeAll * 1000, viewAll)));
        if (CollectionUtils.isNotEmpty(midActionList)) {
            List<String> cidList = new ArrayList<>();
            List<String> converList = new ArrayList<>();
            for (TupleMapEntry<Tuple5> tupleMapEntry : midActionList) {
                String cid = tupleMapEntry.key;
                String click = tupleMapEntry.value.f2;
                String conver = tupleMapEntry.value.f3;
                if (Objects.equals(click, "1")) {
                    cidList.add(cid);
                }
                if (Objects.equals(conver, "1")) {
                    converList.add(conver);
                }
            }
            featureMap.put("user_cid_click_list", String.join(",", cidList));
            featureMap.put("user_cid_conver_list", String.join(",", converList));
        }
        return midActionList;
    }

    private void handleH1AndH2Feature(Map<String, Map<String, String>> skuFeature,
                                      Map<String, Map<String, String>> adVerFeature,
                                      Map<String, String> cidFeatureMap) {
        Map<String, String> h1Feature = adVerFeature.getOrDefault("alg_mid_feature_adver_action", new HashMap<>());
        Map<String, String> h2Feature = skuFeature.getOrDefault("alg_mid_feature_sku_action", new HashMap<>());
        List<String> timeList = Arrays.asList("3d", "7d", "30d");
        List<Tuple2<Map<String, String>, String>> featureList = Arrays.asList(
                new Tuple2<>(h1Feature, "adverid"),
                new Tuple2<>(h2Feature, "skuid")
        );
        for (Tuple2<Map<String, String>, String> tuple2 : featureList) {
            Map<String, String> feature = tuple2.f1;
            String prefix = tuple2.f2;
            for (String time : timeList) {
                String timeValue = feature.getOrDefault(time, "");
                if (StringUtils.isNotEmpty(timeValue)) {
                    String[] split = timeValue.split(",");
                    cidFeatureMap.put("user" + "_" + prefix + "_" + "view" + "_" + time, split[0]);
                    cidFeatureMap.put("user" + "_" + prefix + "_" + "click" + "_" + time, split[1]);
                    cidFeatureMap.put("user" + "_" + prefix + "_" + "conver" + "_" + time, split[2]);
                }
            }
        }


    }

    private void readBucketFile() {
        if (MapUtils.isNotEmpty(bucketsMap)) {
            return;
        }
        synchronized (this) {
            String bucketFile = "20250217_ad_bucket_688.txt";
            InputStream resourceStream = this.getClass().getClassLoader().getResourceAsStream(bucketFile);
            if (resourceStream != null) {
                try (BufferedReader reader = new BufferedReader(new InputStreamReader(resourceStream))) {
                    Map<String, double[]> bucketsMap = new HashMap<>();
                    Map<String, Double> bucketsLen = new HashMap<>();
                    String line;
                    while ((line = reader.readLine()) != null) {
                        // 替换空格和换行符，过滤空行
                        line = line.replace(" ", "").replaceAll("\n", "");
                        if (!line.isEmpty()) {
                            String[] rList = line.split("\t");
                            if (rList.length == 3) {
                                String key = rList[0];
                                double value1 = Double.parseDouble(rList[1]);
                                bucketsLen.put(key, value1);
                                double[] value2 = Arrays.stream(rList[2].split(",")).mapToDouble(Double::valueOf).toArray();
                                bucketsMap.put(key, value2);
                            }
                        }
                    }
                    this.bucketsMap = bucketsMap;
                    this.bucketsLen = bucketsLen;
                } catch (IOException e) {
                    log.error("something is wrong in parse bucket file: ", e);
                }
                log.info("load bucket file success: {}", bucketFile);
            } else {
                log.error("no bucket file");
            }
        }
    }

    private void initSparseFeatureNames() {
        this.sparseFeatureSet = new HashSet<String>() {{
            add("brand");
            add("region");
            add("city");
            add("vid");
            add("cate1");
            add("cate2");
            add("cid");
            add("adid");
            add("adverid");
            add("user_cid_click_list");
            add("user_cid_conver_list");
            add("user_vid_return_tags_2h");
            add("user_vid_return_tags_1d");
            add("user_vid_return_tags_3d");
            add("user_vid_return_tags_7d");
            add("user_vid_return_tags_14d");
            add("apptype");
            add("hour");
            add("hour_quarter");
            add("root_source_scene");
            add("root_source_channel");
            add("is_first_layer");
            add("title_split");
            add("profession");
            add("user_vid_share_tags_1d");
            add("user_vid_share_tags_14d");
            add("user_vid_return_cate1_14d");
            add("user_vid_return_cate2_14d");
            add("user_vid_share_cate1_14d");
            add("user_vid_share_cate2_14d");
            add("user_has_conver_1y");
            add("user_adverid_view_3d");
            add("user_adverid_click_3d");
            add("user_adverid_conver_3d");
            add("user_adverid_view_7d");
            add("user_adverid_click_7d");
            add("user_adverid_conver_7d");
            add("user_adverid_view_30d");
            add("user_adverid_click_30d");
            add("user_adverid_conver_30d");
            add("user_skuid_view_3d");
            add("user_skuid_click_3d");
            add("user_skuid_conver_3d");
            add("user_skuid_view_7d");
            add("user_skuid_click_7d");
            add("user_skuid_conver_7d");
            add("user_skuid_view_30d");
            add("user_skuid_click_30d");
            add("user_skuid_conver_30d");
        }};
    }

    private Map<String, String> featureBucket(Map<String, String> featureMap) {
        Map<String, String> newFeatureMap = new ConcurrentHashMap<>(featureMap.size());
        for (Map.Entry<String, String> entry : featureMap.entrySet()) {
            String name = entry.getKey();
            if (this.sparseFeatureSet.contains(name)) {
                if (entry.getValue() != null) {
                    newFeatureMap.put(name, entry.getValue());
                }
                continue;
            }
            double score = Double.parseDouble(entry.getValue());
            // 注意：0值、不在分桶文件中的特征，会被过滤掉。
            if (score > 1E-8) {
                if (this.bucketsMap.containsKey(name) && this.bucketsLen.containsKey(name)) {
                    double[] buckets = this.bucketsMap.get(name);
                    double bucketNum = this.bucketsLen.get(name);
                    Double scoreNew = 1.0 / bucketNum * (ExtractorUtils.findInsertPosition(buckets, score) + 1.0);
                    newFeatureMap.put(name, String.valueOf(scoreNew));
                } else {
                    newFeatureMap.put(name, String.valueOf(score));
                }
            }
        }
        return newFeatureMap;
    }
}
